#if 0
C {0}.o interpreter.o ::obj:disasm6502.c.o ::obj:ftrig.a.o -o asm6502 -d SAMPLE -d CHR_HX=36 -d PS=58 -d ALLOW_BIN_HI_LO -d NO_EBREAKS -d ALLOW_DLITERALS -d ALLOW_NAME_DOT {"Parameters"}
Exit 0
#endif
/*
 asm6502.c
  .shl (<<), .shr (>>), .byte, .word, .localchar (? ? ?)
   and .segment (externally defined range) are not supported.
  only .include preprocessing is implemented (macros are not).

  Revised: 2013.05.29

 * Copyright notice for this file:
 * asm6502.c Copyright (C) 2010, 2013 adamjasonp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


2013.05.19
 Got saw_song.asm to compile and run correctly
  - saved 52 bytes with optimization feature on (from command-line):
  colors2.asm -Ccolors2.chr -M -O -Ncolors2.nes
11 AM
 Eventually got colors2.nes to work - holy crap, the exact same results as
  what Telemark Cross Assembler 3.2 and MAKEINES produced!
2013.05.20
 Now with modules implemented and non-indexed indirect mode parsing fixed,
  all of the basics are complete.  Now, if only FDSPic would work...
2013.05.21
 Added PRG floor (for custom file output).
2013.05.22
 Fixed the nature of several things...many bug fixes, feature changes.
 Added "logical smallest" for zero PRG input.
 More syntax/error verification, eliminating atoi (-M needs a value now).
 Implemented sections: position swapping (section org saved), clipping
  (for zero-page - a mask to clear the high byte), no writing in BSS/ZP.
2013.05.23
 More range checking, bug fixes; added PWrError and use.
 Rewrote textlooping to write directly instead of malloc...
 AsmFile: labels can now be on text-include lines.
 ORG now also seconds as PAD if a second parameter is present.
 X6502_Assemble, now more professional, takes s-e range, error on site.
 getOpcode (formerly getInstruction) now takes BAD flag in mind - in
  other words, skip the ones we don't want encoded.
 Fixed entries in x6502_opn_table; ROL/ROR w/o A operand were going to
  undefined "?OPs" (undefined, no-operation codes, such as $7A).

2013.05.24
 On account of mmc5exram.asm, it appears that ORG/SPS swapping was wrong.
 The actual position written does _not_ come directly from ORG.
 Done.  Results are now identical with:
  mmc5exram.asm -Cmmc5exram.chr -M5 -R1 -O
 Though starting in the code section by default, ORG is initialized to
  zero to work with the V-mir. scanline.asm (-Cscanline.chr -M0 -R1 -O).
 Added submapper, maxing out the 8 option bits.
 Implemented .lsb and .msb.
 Added _orgInited.

2013.05.25
 Fixed bank writing bug that wrapped all output to lower 32KB.
 PRG can now be lower than 16KB for iNES (repeated output).
 Fixed floor command line input.
 Got logical minimum working.
 Instructions with inapplicable addressing modes didn't err w. opt off
  - fixed.

2013.05.26
 Added bank masking (and 'flat' option), fixing bank management.
 Added include path parent concatenation.
 Fixed parameter advance for .opt and .incbin.
 To work with zpong3 again, with .code, ORG is set to pos if <$2000;
  pos is adjusted on .org in code section and ORG set.
 Labels/equates are back to being 32-bit.
 Wrote label scanning ('-' and '+') w. length as count; matched PALTEST.

2013.05.27
 Auto-bankswitch increments on $C000 regardless of new ORG being lower.
 Added .addr (dc.w), .byt (dc.b) and .res (block).
 Added ':' skipping for label operands.
 Added .endproc and .proc; they currently act as "module subroutines."
 Error return wasn't cleared on empty string - fixed.
 Module names affect labels with '@', not just '_'.
 Added some additional error/warning codes.
 Fixed open mode in incbin (was "r"), which subsequently fixed tvpassfail
  (with .section conversion to flat, reordering procs until identical).

2013.05.29
 Added PRG/CHR RAM/WRAN options, checks to force version two.
  The setting of two fields fixed in detection.
 Fixed five op table codes that allowed an invalid immediate operand:
  STY/X/A, DEC and INC.


Notes:
 PALTEST and some other carts call for flat addressing (-L):
  "PALTEST.ASM -P32 -L -M1 -O"

 The NEStress assembly source (unknown credit) appears to be rather stale,
  2002-07-23 to the 2002-07-31 build.  The source file has text shows a
  score of 9 instead of 11.  The .nes hangs in the IRQ routine.
  "-CTanks.CHR -P32 -L -M0 -R1 -O"

 FDSPic.asm is really 8KB (-P8 -M3 -O -CFDSPic.chr -R1).
*/

/*
 Preprocessing considerations:
	.macro <name> [<param1>, [<param2> [...]]]
	.endmacro
	.if <expr>
	.else
	.endif

 DrPCjr appears to have a 512 byte header.
*/

#include "../enviro64.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "interpreter.h"
#include "ftrig.h"

#include "asm6502.h"

/* ---- macros */

#define IS_EOL(c)		((uint8)((c)-10u)<=(uint8)3u)
#define SKIPCHR(s,e,c)		do{while((s)<e&&*(s)==(c)) ++(s);}while(0)

/* ---- code */

static const char *IDBreak(const char *str,
	const char *e)
{
	const char	*s = str;

	if (s>=e)
		return s;
	if (*s=='@') /* one starting 'at' only */
		if (++s>=e)
			return str;
	do
		switch(*s)
		{
		case '.': /* only one consecutive '.' */
			//if (s>str && s[-1]=='.')
			//	return s;
		case '_':
		CASE_NUMERIC
		CASE_ALPHABETIC
		next:	++s;
			break;

		default:
#ifdef CHR_NB
			if (*s==CHR_NB)
				goto next;
#endif
			return s;
		}
	while(s<e);

	return s;
}

static int appendParent(char **b,
	const char *pfn)
{
	const char	*n;
	char		*p;
	long		x, z;

	if (!b || (z=strlen(p=*b))<0)
		return _e_general_error;
#if PS=='\\' /* Windows */
	if (*pfn==':' || (pfn[1] && pfn[1]==':'))
#elif PS==':' /* Mac OS */
	if (*pfn!=PS)
#else
	if (*pfn==PS)
#endif
		goto end; /* explicit path */
	if ((n=strrchr(pfn,PS))==NULL || (x=n-pfn)<=0)
		goto end;
	if (!z++ || p[z-2]!=PS)
		++x;
	/* insert before */
	if ((p=realloc(p,x+z))==NULL)
		return _e_memory_error;
	*b = p;
	memmove(p+x,p,z);
	memmove(p,pfn,x);
end:	return 1;
}

long X6502_ParseName(const char *str,
	const char *e,
	long *z,
	char **b,
	const char *parent) /* added 2013.05.26 */
{
	const char	*s;
	char		*p;
	long		r;

	if (!str || str>=e)
		return _e_general_error;
	if (*str=='\047' || *str=='"')
	{
		r = IMAX(r);
		s = parse_textloop(str,e,&r,NULL,b,0,NULL);
		if (s>=e)
			return _e_open_string;
		if (*s++!=*str) /* equal close quote or point of error */
			return _e_general_error;
		e = NULL; /* buffered - not directly from str */
	} else /* can expect an extension */
		r = (s=IDBreak(str,e))-str;
	if (r<0)
		return 0;
	if ((p=realloc(*b,r+1))==NULL)
	{
		if (*b)
			{free(*b); *b = NULL;}
		return _e_memory_error;
	}
	(*b=p)[r] = 0; /* make sure this is C-string terminated */
	if (z)
		*z = r;
	if (e)
		memcpy(*b,str,r);
	if (parent && (r=appendParent(b,parent))<0)
		return r;
	return s-str;
}

static int labelScan(asm6502frame *f,
	evalvalue *v,
	int c,
	uint32 o)
{
	const uint8	*s, *l;
	int32		t =-1;

	if (!f || (s=f->lab)==0)
		return _e_general_error;
	if (!*(l=s))
		return 0;
	/* scan forward first (reverse after getting to the closest ORG) */
	for(; *l++; l+=1+4)
		if ((*(l+=*l+2)&0xBF)==_Label) /* must be an established label */
			if ((t=l[1]|((l[2]|((l[3]|(l[4]<<8))<<8))<<8))>=o)
				if (c<0)
				{
					if (!(c+=o>=t))
						goto val;
					goto rev;
				} else if ((c-=t>o)<=0)
					goto val;
	if (c<0)
	{
		l-= 1+1+4;
		do
		{
			if ((*l&0xBF)==_Label && !++c)
			{
				t = l[1]|((l[2]|((l[3]|(l[4]<<8))<<8))<<8);
			val:	if (v)
					*v = t;
				return 1;
			}
		rev:	;
		} while((l-=l[-1]+2+4+2)>s);
	}
	return 0;
}

static const char *parseValue(const char *s,
	const char *e,
	evalvalue *i,
	void *u)
{
	asm6502frame	*f;
	uint8		b[256], *d;
	const char	*n;
	int32		x;
	int		z, h;

	if (s>=e)
		return NULL;
	f = u;
	if (*s=='*')
	{ /* current location */
		if (!f)
			return NULL;
		if (i)
			*i = f->org;
		return s+1;
	} else if (*s=='%')
	{ /* binary */
		uint32		v = 0;
		uint8		c;

		if (++s>=e)
			return NULL;
		do
			if ((c=*s-'0')>1)
			{ /* can have no other following alphanumerics */
				if (c<=9 || (c+='0'-'A')<='Z'-'A'
						|| (c+='A'-'a')<='z'-'a')
					s = NULL;
				break;
			} else
				v+= v, v+= c;
		while(++s<e);
		if (i)
			*i = v;
		return s;
	}
	n = IDBreak(s,e), h =-1;
	if (toupper(*s)=='A' && n==s+1)
		return NULL; /* register-reserved */
	if (s==n)
	{
		if (*s=='-')
			{SKIPCHR(n,e,'-'); x = s-n;}
		else if (*s=='+')
			{SKIPCHR(n,e,'+'); x = n-s, f->flg|= _PreEstID;}
		else
			return NULL;
		if (labelScan(f,i,x,(uint32)f->org)<=0)
			if (!(f->opt&_Prescan))
				{f->flg|= _LScanErr; return NULL;}
			else if (i)
				*i = f->org+3;
		return n;
	} else
	{
		e = n;
		if (n-s>4 && n[-4]=='.'
				&& toupper(n[-1])=='B' /* BS */
				&& toupper(n[-2])=='S')
		{ /* (somewhat inefficient) handling of .LSB/.MSB */
			uint8		c = toupper(n[-3]);

			if (c=='L')
				h = 0, e-= 4;
			else if (c=='M')
				h = 8, e-= 4;
		}
	}
	if ((x=X6502_AsmGetID(s,e,b,u,&z))>0)
	{
		d = ((uint8*)f->lab)+x;
		if (!(f->opt&_Prescan) && (*d&0x3F)==_Unestablished)
			return NULL; /* should be defined by now */
	} else if (X6502_ParseKey(s,e,_off)>=0 /* can't be a keyword; 2013.05.21 */
			|| !(f->opt&_Prescan)
			|| (x=X6502_AsmAddID(s,e,b,u,_Unestablished|0x80))<=0)
		return NULL;
	else
	{
		d = ((uint8*)f->lab)+x;
		d[2] = f->org>>8; /* page copy */
	}
	if (*d&0x80)
		f->flg|= _PreEstID;
	if (i)
	{
		x = d[1]+(d[2]<<8);
		if (h>=0)
			*i = x>>h&0xFF;
		else
			*i = x+((d[3]+(d[4]<<8))<<16);
	}
	return n;
}

#if 0
static char	modes[][3] =
{
	"IM", /* 1-immediate		- #$xx */
	"RL", /* 1-relative		- *+$xx */
	"ZP", /* 1-zero page		- $xx */
	"ZX", /* 1-zero page, X index	- $xx,X */
	"ZY", /* 1-zero page, Y index	- $xx,Y */
	"IX", /* 1-indirect, X index	- ($xx,X) */
	"IY", /* 1-indirect, Y index	- ($xx),Y */
	"IN", /* 1-indirect		- ($xx) */
	"IB", /* 2-absolute indirect	- ($xxxx) */
	"AB", /* 2-absolute		- $xxxx */
	"AX", /* 2-absolute, X index	- $xxxx,X */
	"AY", /* 2-absolute, Y index	- $xxxx,Y */
	"IA", /* 0-implied A to self (e.g., LSR) */
	"IP"
};
#endif

/* returns the total parsed length, and larger operand mode in `m` */
long X6502_ParseOperands(const char *str,
	const char *e,
	evalvalue *v,
	asm6502frame *f,
	int *m)
{
	const char	*s = str, *n;
	long		r;
	int		t;
	char		b = 0, c;

	if (s>=e || (c=*s)==';' || IS_EOL(c))
		{*m = OPM_IP; return 0;}
	if ((n=memchr(s,';',e-s))==NULL)
		n = e;
	else
		e = n;
	if (c=='#')
		*m = OPM_IM, ++s; /* immediate */
	else if (c=='(')
	{ /* indirect - must scan for each evaluating bounds */
		b = ')';
	ib:	*m = OPM_IB, n = ++s, r = 1;
		for(; n<e && (c=*n)!=','; ++n)
			if (c=='(')
				++r;
			else if ((c==')' || c==']') && !--r)
				break;
	} else if (c=='[')
		{b = ']'; goto ib;}
	else
		*m = OPM_AB; /* absolute (or ZP if low and opt on) */
	if ((r=evaluate_line(s,n,v,&t,parseValue,NULL,f,NULL))==0 || t)
	{ /* for pre-labels - you may get away with `[-]` */
		if (*s==':')
			++s; /* for ca65 `:-`, `:+`, etc. */
		if ((n=parseValue(s,n,v,f))==NULL)
		{
			if (parse_skipwsp(str+1,e)>=e && toupper(*str)=='A')
				{*m = OPM_IA; return e-s;} /* A register */
			return t<0 ? t==_e_incomplete_error && f->flg&_LScanErr
					? _e_labelscan_overflow:t
					:_e_general_error;
		} else
			r = n-s;
	}
	s = parse_skipwsp(s+r,e);
	r = e-str;
	if (s>=e /*|| *s==';'*/ || IS_EOL(*s))
		return b ? _e_general_error:r;
	if (*s==')' || *s==']')
	{ /* bracket/parenthesis must match, IB or IY, resulting */
		if (!b || *s!=b || *m!=OPM_IB)
			return _e_general_error;
		t = 1, s = parse_skipwsp(s+1,e);
	}
	if (*s==',')
	{
		s = parse_skipwsp(s+1,e);
		if (s>=e) /* only a comma? */
			return _e_general_error;
		if ((c=toupper(*s))=='X' && !t)
			if (*m==OPM_AB)
				*m = OPM_AX;
			else if (*m==OPM_IB)
				*m = OPM_IX;
			else
				return _e_general_error;//_e_op_error;
		else if (c=='Y')
		{ /* must be closed first */
			if (*m==OPM_AB)
				*m = OPM_AY;
			else if (*m==OPM_IB && t)
				*m = OPM_IY;
			else
				return _e_general_error;//_e_op_error;
		} else
			return _e_general_error;
		s = parse_skipwsp(s+1,e);
		/* need indirect closing?  IY doesn't. */
		if (s>=e || IS_EOL(*s) /*|| *s==';'*/)
			return *m==OPM_IX ? _e_general_error:r;
		if (*s!=b) /* IX does */
			return _e_general_error;
		s = parse_skipwsp(s+1,e);
	} else if (b && !t) /* need indirect closing */
		return _e_general_error;
	if (s<e && !IS_EOL(*s) /*&& *s!=';'*/)
		return _e_general_error;
	return r;
}

/* ---- keywords */

static const char	*keyw_list[] =
{
	"INCBIN",
	"INCLUDE",
	"ADDR",
	"ALIGN",
	"BANK",
	"BLOCK",
	"BSS",
	"BYT",
	"CODE",
	"DB",
	"DCB",
	"DC.B",
	"DEFAULT",
	"DEFCHR",
	"DW",
	"DCW",
	"DC.W",
	"END",
	"ENDPROC",
	"EQU",
	"FILL",
	"INDEX",
	"INESCHR",
	"INESMAP",
	"INESMIR",
	"INESPRG",
	"MEM",
	"MODULE",
	"OPT",
	"ORG",
	"P02",
	"PAD",
	"PROC",
	"PROCESSOR",
	"RES",
	"SEGMENT",
	"SUBROUTINE",
	"ZEROPAGE",
	"ZP",
	"OFF",
	"ON"
};

#define NUMBER_OF_KEYS		(OPN_COUNT+GET_ARRAY_COUNT(keyw_list))

int X6502_ParseKey(const char *s,
	const char *e,
	uint32 max)
{
	char		n[10];
	int		i, z = e-s;
	uint8		m = 0;

	if (!s || s>=e)
		return _e_general_error;
	if (max>NUMBER_OF_KEYS)
		max = NUMBER_OF_KEYS;
	if (*s=='.')
		++s, m = 1;
	if ((z=e-s)>10)
		goto nope;
	i = 0;
	do
		n[i] = toupper(*s++);
	while(++i,s<e);

	if (z==3 && !m)
	{
		const char	*t = *x6502_opnametable;

		i = 0;
		do
			if (*t==*n && t[1]==n[1] && t[2]==n[2])
				return i;
		while(t+=3,++i<OPN_COUNT);
	} else
		i = OPN_COUNT;
	if (max
#ifdef KEYW_REQUIRES_DOT
			&& m
#endif
			)
	{ /* keyword (e.g., .equ) */
		const char	**t = keyw_list, *s, c = *n;

		--z;
		do
			if (*(s=*t)==c && !memcmp(++s,n+1,z) && !s[z])
				return i;
		while(++t,++i<max);
	}
nope:	return _e_no_identifier;
}

/* returns ID type/data offset, ~end or zero on error */
int32 X6502_AsmGetID(const char *s,
	const char *e,
	uint8 *n,
	const asm6502frame *f,
	int *_z)
{
	const uint8	*l;
	uint8		h, z;

	if (!f || !n || !s || e<=s)
		return 0;
	*n = toupper(*s++)&0x3F, z = h = 0;
	if (f->mod)
		if (*n==('_'&0x3F) || *n==('@'&0x3F))
		{
			long	x;

			if ((x=strlen(f->mod))>255)
				return 0; /* no room for anything else? */
			memcpy(n+1,f->mod,x);
			z = x, *n = (*n&1)|0x20; /* '@' to ' ', '_' to '!' */
		}
	while(s<e)
		if (!++z)
			return 0;
		else
			h+= n[z] = toupper(*s++);
	if ((h=h<<6|*n)==0) /* '@' is valid */
		h = 0x80; /* hash can't be null */
	*n++ = h;
	if (_z)
		*_z = 1+z;
	if ((l=f->lab)==0)
		return ~0;
	/* hash and type match */
	for(; *l++; l+=*l,l+=1+2+4)
		if (*l==z && h==l[-1] && !memcmp(l+1,n,z))
			return 2+z+l-f->lab;
	return f->lab-l; /* becomes ~total */
}

/* returns ID offset, or <=0 on error */
int X6502_AsmAddID(const char *s,
	const char *e,
	uint8 *n,
	asm6502frame *f,
	uint8 t)
{
	uint8		*d;
	int32		x;
	int		z;

	if ((x=X6502_AsmGetID(s,e,n,f,&z))>=0)
	{
		if (!x)
			return 0;
		d = f->lab+x;
		if (t!=_Unestablished && !(*d&0x40) && (*d&0x3F)!=_Unestablished)
			return -x;
		if (f->opt&_Prescan)
			t|= *d&0x80; /* preserve PreEst */
		goto typ; /* can only establish once */
	}
	/* (-x) end offset including null term., 1+1 length, 1 typ, 4 val */
	if ((d=realloc(f->lab,z+1+2+4-x))==0)
		return _e_memory_error;
	f->lab = d, d-= x;
	d[-1] = *n++; /* hash (includes first char) */
	*d++ = x = z-1; /* post-hash length */
	memcpy(d,n,x); /* post-hash name */
	d+= x, *d++ = x; /* duplicate-length for reverse scanning */
	d[1] = d[2] = d[3] = d[4] = d[5] = 0; /* init and terminate */
typ:	if (f->opt&_Prescan)
		t|= 0x40;
	*d = t;
	return d-f->lab;
}

/* ---- assembler */

static int getOpcode(uint16 t)
{
	int		i = 0;

	do /* mask used OPB_BAD - BAD. */
		if ((x6502_opn_table[i]&(OPB_FLAG-1))==t)
			return i;
	while(++i<256);

	return ~0;
}

/* returns source text name length or error */
static long incbin(const char *str,
	const char *e,
	asm6502frame *f,
	const char *fn)
{
	void (*bw)(asm6502frame*,
		uint8);
	FILE		*nf = NULL;
	char		*b = NULL;
	long		z;

	if ((z=X6502_ParseName(str,e,NULL,&b,fn))>0)
		nf = fopen(b,"rb");
	if (b)
		free(b);
	if (!nf)
		return z ? (z>0 ? _e_include_open:z):_e_incomplete_error;
	if ((bw=f->writef)==NULL)
	{
		fseek(nf,0,SEEK_END);
		f->org+= ftell(nf);
	} else
	{
		int		c;

		for(; (c=fgetc(nf))>=0; ++f->org)
			bw(f,c);
	}
	fclose(nf);
	return z;
}

static void incchar(void *u,
	char c)
{
	asm6502frame	*f;

	if ((f=u)==NULL)
		return;
	if (f->writef)
		f->writef(f,c);
	++f->org;
}

static int optimizeMode(int m,
	uint32 v)
{
	if (v>255)
	{ /* expand */
		if (m==OPM_ZP)
			m = OPM_AB;
		else if (m==OPM_IN)
			m = OPM_IB;
		else if (m==OPM_ZX)
			m = OPM_AX;
		else if (m==OPM_ZY)
			m = OPM_AY;
	} else if (m==OPM_AB) /* reduce */
		m = OPM_ZP;
	else if (m==OPM_IB)
		m = OPM_IN;
	else if (m==OPM_AX)
		m = OPM_ZX;
	else if (m==OPM_AY)
		m = OPM_ZY;
	return m;
}

static int makeTmpLabelName(uint8 *d,
	asm6502frame *f)
{
	if (!d || !f)
		return _e_general_error;
	*d = ':';
	return makebasestring(d+1,NULL,f->org|f->bnk<<16,16,6,0)+1;
}

#define prepWr(f,r)	do{ \
	if (f->sct!=_codeSection) \
		{r = _e_write_in_addr_sect; goto end;} \
	if (!(f->flg&_orgInited)) \
		f->flg|= _orgInited, f->org = f->pos; \
}while(0)

long X6502_Assemble(const char *str,
	const char *e,
	int *_r,
	asm6502frame *f,
	const char *fn)
{
	uint8		b[256];
	asm6502frame	af;
	void (*bw)(asm6502frame*,
		uint8);
	const char	*s;
	int		r, m, wrn = 0;
	uint16		k;
	uint8		c = 0;

	if ((s=str)==NULL)// || s>=e)
		return 0; /* empty */
	if (!f) /* scratch - clear */
		memset(f=&af,0,sizeof(af));
	r = 0, f->flg&=~(_PreEstID|_LScanErr), bw = f->writef;
	k = f->sct==_zpSection ? 0xFF:0xFFFF;
	while(s<e)
	{
		evalvalue	v;
		const char	*n, *le, *ls;
		long		l, lorg;
		int		x;

		if (IS_EOL(*s)) /* new line */
			{f->flg&=~_LineFlags, ++s; continue;}
		if (c)
			{++s; continue;}
		if (*(n=ls=parse_skipwsp(s,e))==';')
			{c = 1, s = n+1; continue;}
		if (n!=s)
			f->flg|= _IntoLine;
		m =-1, lorg = f->org, n = IDBreak(s=n,e), le = e;
		if ((x=X6502_ParseKey(s,n,_off))>=0)
		{
			/* some assemblers allow non-instruction keys at start */
			if (x<OPN_COUNT && !(f->flg&_IntoLine))
				{r = _e_label_error; goto end;}
			if (x==_default || x==_proc)
			{
				s = parse_skipwsp(n,e), n = IDBreak(s,e);
				goto labo;
			} else
				goto gotn;
		}
		if (*n==':' || !(f->flg&_IntoLine))
		{ /* identifier: equate/label */
			long		z;
			uint8		t;

		labo:	t = _Label;
			if ((z=n-s)==0) /* '-' mark (pre-)labels */
			{
				if (*n==':')
					z = 1; /* will be automatically skipped */
				else
					SKIPCHR(n,e,'-');
				z = makeTmpLabelName(b+128,f), s = (char*)b+128;
			} else if (x==_default)
				t = _Default;
			if (toupper(*s)=='A' && z==1)
				{r = _e_identifier_conflict; break;} /* register-reserved */
			if ((m=X6502_AsmAddID(s,s+z,b,f,t))<=0)
				{r = m ? _e_identifier_conflict:_e_name_too_long; break;}
			if (x==_proc) /* put here since trailing ':' is invalid */
				goto gotp; /* reuse for module name (indep. labels) */
			if (*n==':')
				++n;
		}
		if (*(n=parse_skipwsp(n,e))==';')
			le = n;
		s = n;
		if ((n=IDBreak(n,e))<=s)
		{
			if (*s=='=') /* '=' equate */
			{
				++n;
				if (x!=_default)
				{
					if (m<0)
						{r = _e_general_error, s = str; break;}
					x = _equ;
				}
				goto gotn;
			}
			if (m>=0)
				goto lset;
			if (x<0 && ls<le)
				s = ls, r = _e_macro_error;
			break;
		} else if (x==_default) /* must follow '=' */
			{r = _e_general_error; break;}
		/* macro or instruction */
		if ((x=X6502_ParseKey(s,n,_off))<0)
			{r = x==_e_no_identifier ? _e_macro_error:x; break;}
		/* keyword, possible equate, etc. */
	gotn:	s = parse_skipwsp(n,e);
	gotp:	r = 0; /* now at position of the parameter(s) */
		switch(x)
		{ /* the `r` end-value when using lset displaces org */
		case _align:
			s+= evaluate_line(s,e,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return && (r!=_e_general_error || s>=e || *s!=';'))
				goto end;
			if (v-1u>0xFFFFu-1u)
				{r = _e_range_error; goto end;}
			f->org+= v-1, f->org-= f->org%v, f->org&= k, r = 0;
			if (f->sct==_codeSection)
			{
				if ((l=f->org-lorg)<0)
					wrn =~_e_code_wrap;
				f->pos+= l;
			}
			break;
		case _bank: /* is this one-based? */
			s+= evaluate_line(s,e,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return && (r!=_e_general_error || s>=e || *s!=';'))
				goto end;
			/* 32 KB always mapped, even if mirrored */
			if ((unsigned)v>f->maxprg && (unsigned)v>1)
				{r = _e_range_error; goto end;}
			f->bnk = v, r = 0;
			break;
		case _block: /* asign block, skip (add to ORG) */
		case _res:
			s+= evaluate_line(s,e,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return && (r!=_e_general_error || s>=e || *s!=';'))
				goto end;
			f->org+= v, f->org&= k;
			goto lset;
		case _bss: /* RAM and I/O addressing */
			//if (f->sct==_codeSection)
			//	f->org = f->sps, f->sps = lorg; /* swap */
			f->sct = _bssSection, k = 0xFFFF; /* may overflow */
			break; /* cannot put a label on this line */
		case _code:
			//if (f->sct!=_codeSection)
			//	f->org = f->sps, f->sps = lorg; /* swap */
			f->sct = _codeSection;
			if (!(f->flg&_orgInited) || f->org<0x2000)
				f->flg|= _orgInited, f->org = f->pos;
			k = 0xFFFF;
			break;
		case _byt:
		case _db:
		case _dcb:
		case _dc_b:
			prepWr(f,r);
			x = 0;
			do
			{
				if (x && *s++!=',')
					{r = _e_general_error; goto end;}
				s = parse_skipwsp(s,e);
				if (*s=='\047' || *s=='"')
				{
					l = IMAX(r), n = s;
					s = parse_textloop(s,e,&l,NULL,NULL,f,incchar);
					if (s>=e)
						{r = _e_open_string; goto end;}
					if (*s++!=*n) /* equal close quote or point of error */
						{r = _e_general_error; goto end;}
				} else
				{
					s+= evaluate_line(s,e,&v,&r,parseValue,NULL,f,NULL);
					if ((unsigned)r>=_e_return && (r!=_e_general_error
							|| s>=e || *s!=';'))
						goto end;
					if (v>255 || v<-0x80)
						wrn =~_e_range_error;
					incchar(f,v);
				}
			} while(++x,(s=parse_skipwsp(s,e))<e
					&& !IS_EOL(*s) && *s!=';');
			x = _db;
			goto lset;
		case _defchr:
			prepWr(f,r);
			l = 0;
			do
			{ /* supposed to pick up 8 hex codes, proc. 2 bits of ea. nybble */
				uint16	dp;

				if (l && ((s=parse_skipwsp(s,e))>=e || *s++!=','))
					{r = _e_general_error; goto end;}
				s+= evaluate_line(s,e,&v,&r,parseValue,NULL,f,NULL);
				if ((unsigned)r>=_e_return && (r!=_e_general_error || s>=e || *s!=';'))
					goto end;
				dp = 0, x = 8;
				do
					dp+=dp, dp|= (v>>28-8)&0x100, dp|= (v>>28+1)&1;
				while(v<<=4,--x);
				b[l] = dp>>8, b[l+8] = dp;
			} while(++l<8);
			if (!bw)
				f->org+= 16;
			else
			{
				x = 0;
				do
					bw(f,b[x]);
				while(++f->org,++x<16);
			}
			x = _defchr;
			goto lset;
		case _addr:
		case _dw:
		case _dcw:
		case _dc_w:
			prepWr(f,r);
			l = 0;
			do
			{
				if (l && *s++!=',')
					{r = _e_general_error; goto end;}
				s+= evaluate_line(s,e,&v,&r,parseValue,NULL,f,NULL);
				if ((unsigned)r>=_e_return && (r!=_e_general_error || s>=e || *s!=';'))
					goto end;
				if (v>65535 || v<-0x8000)
					wrn =~_e_range_error;
				if (!bw)
					f->org+= 2;
				else
				{
					bw(f,v), ++f->org;
					bw(f,v>>8), ++f->org;
				}
			} while(++l,(s=parse_skipwsp(s,e))<e
					&& !IS_EOL(*s) && *s!=';');
			goto lset;
		case _end:
			if (f->flg&_IntoProc)
				wrn =~_e_open_proc;
			f->flg|= _EndKeyw;
			break; /* cannot put a label on 'end' lines */
		case _endproc:
			if (f->flg&_IntoProc)
			{
				f->flg&=~_IntoProc;
				if (f->mod)
					{free(f->mod); f->mod = NULL;}
			} else
				wrn =~_e_no_proc;
			break; /* cannot put a label on 'end' lines */
		case _default:
		case _equ: /* set label to explicit value */
			s+= evaluate_line(s,e,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return && (r!=_e_general_error || s>=e || *s!=';'))
				goto end;
			if (m<0)
				{r = _e_no_identifier; goto end;}
			if (f->lab)
			{
				r = x==_default ? _Default:_Equate;
				if (f->opt&_Prescan)
					r|= 0x40;
				f->lab[m] = r;
			}
			lorg = v;
			goto lset;
		case _incbin:
			prepWr(f,r);
			if ((r=incbin(s,e,f,fn))<=0)
				goto end;
			s+= r;
			goto lset;
		//case _off: /* on/off should not be used as labels */
		//case _on:
		case _include: /* read this file in place of line */
			r = _e_include; //_e_label_error;
			goto lset; /* make sure any label is set */
		case _index:
			wrn =~_e_handler_error;
			s+= evaluate_line(s,e,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return && (r!=_e_general_error || s>=e || *s!=';'))
				goto end;
			r = 0;
			break;
		case _ineschr:
			s+= evaluate_line(s,e,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return && (r!=_e_general_error || s>=e || *s!=';'))
				goto end;
			if ((unsigned)v>512)
				{r = _e_range_error; goto end;}
			if (f->maxchr>=0 && f->maxchr!=v)
				{r = _e_value_set; goto end;}
			f->maxchr = v, r = 0;
			break;
		case _inesmap:
			s+= evaluate_line(s,e,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return && (r!=_e_general_error || s>=e || *s!=';'))
				goto end;
			if ((unsigned)v>512)
				{r = _e_range_error; goto end;}
			if (f->map>=0 && f->map!=v)
				{r = _e_value_set; goto end;}
			f->map = v, r = 0;
			break;
		case _inesmir:
			s+= evaluate_line(s,e,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return && (r!=_e_general_error || s>=e || *s!=';'))
				goto end;
			if ((unsigned)v>2)
				{r = _e_range_error; goto end;}
			r = f->opt&_MirrorMask;
			if (r!=_MirrorMask && r!=v)
				{r = _e_value_set; goto end;}
			f->opt&=~_MirrorMask, f->opt|= v, r = 0;
			break;
		case _inesprg:
		case _mem:
			s+= evaluate_line(s,e,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return && (r!=_e_general_error || s>=e || *s!=';'))
				goto end;
			if (x==_mem)
				v>>= 2; /* a guess */
			if ((unsigned)--v>=(0x3FFFFFF>>f->bsz) || v>=0x7FFFu)
				{r = _e_range_error; goto end;}
			if (f->maxprg>=0 && f->maxprg!=v)
				wrn =~_e_value_set;
			f->maxprg = v, r = 0;
			break;
		case _proc:
			if (f->flg&_IntoProc)
				{r = _e_already_in_proc; goto end;}
			f->flg|= _IntoProc;
			wrn =~_e_cannot_optimize;
		case _module:
			if (IDBreak(s,e)==s /* can't stuff *any* char into labels */
					|| (r=X6502_ParseName(s,e,&l,&f->mod,NULL))==0)
				{r = _e_general_error; goto end;}
			if (r>255) /* overflow */
				r = _e_name_too_long;
			if (r<0)
				goto end;
			s+= r;
			goto lset;
		case _opt:
			r = X6502_ParseKey(s=parse_skipwsp(s,e),e,-1);
			if ((unsigned)(r-=_off)>1)
				{r = _e_off_or_on; goto end;}
			f->opt&=~_Optimize, f->opt|= r*_Optimize, r = 0;
			s = IDBreak(s,e);
			break;
		case _processor:
			if ((n=IDBreak(s,e))==s)
				{r = _e_general_error; goto end;}
			if (n-s!=4 || memcmp(s,"6502",4))
				{r = _e_language_error; goto end;}
			s = n;
			break;
		case _segment:
			r = _e_handler_error;
			goto end;
		case _subroutine:
			wrn =~_e_cannot_optimize;
			goto lset;
		case _zp:
		case _zeropage:
			k = 0xFF;
			//if (f->sct==_codeSection)
			//	f->org = f->sps&k, f->sps = lorg; /* swap */
			f->sct = _zpSection;
			goto lset;
		case _fill: /* fill bytes of length, optional value */
		case _org: /* set position */
		case _pad:  /* pad to this next address */
			s+= evaluate_line(s,e,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return && (r!=_e_general_error || s>=e || *s!=';'))
				goto end;
			if ((unsigned)v>0xFFFF)
				{r = _e_range_error; goto end;}
			if (x==_fill)
				l = v;
			else
				l = v-lorg;
			if (*(s=parse_skipwsp(s,e))==',')
			{
				v = 0, ++s;
				s+= evaluate_line(s,e,&v,&r,parseValue,NULL,f,NULL);
				if ((unsigned)r>=_e_return && (r!=_e_general_error || s>=e || *s!=';'))
					goto end;
			} else if (x==_org)
			{
				if (f->sct==_codeSection && f->flg&_orgInited && v>=0x8000)
					f->pos = v;
				f->org+= l, f->org&= k;
				f->flg|= _orgInited;
				//if (f->opt&_AutoBank)
				if (f->org==0xC000)
				{
					if (f->bnk<0)
						f->bnk = 0;
					if (/*v<f->org &&*/ ++f->bnk>f->maxprg && f->maxprg)
						f->bnk = 1, wrn =~_e_bank_overflow;
					f->pos = f->org;
				} else if (f->org==0x8000) /* lower, fixed bank */
					f->bnk = 0, f->pos = f->org;
				r = 0;
				//pos = f->org;
				//goto lset;
				break;
			} else
				v = 0;
			prepWr(f,r);
			if (l>0)
				if (!bw)
					f->org+= l;
				else
					do
						bw(f,v);
					while(++f->org,--l);

		default: /* label at org */
		lset:	if (m>=0)
			{
				uint8	*d;

				if ((d=f->lab)==NULL)
					{r = _e_read_error; break;}
				d+= m;
				d[1] = lorg;
				d[2] = lorg>>8;
				d[3] = lorg>>16;
				d[4] = lorg>>24;
			}
			if (r!=_e_include)
				r = 0;
			m =-1;
		}
		if (m>=0) /* must be cleared by now */
			{r = _e_label_error; goto end;}
		s = parse_skipwsp(s,e);
		if ((unsigned)x>=OPN_COUNT)
		{ /* not encoding an instruction */
			if (s>=e || IS_EOL(*s))
				{f->flg&=~_LineFlags, ++s; continue;}
			if (*s==';')
				continue;
			if (r!=_e_include)
				r = _e_general_error;
			goto end;
		}
		prepWr(f,r);
		r = X6502_ParseOperands(s,e,&v,f,&m);
		if ((unsigned)r>=_e_return)
			goto end;
		s+= r;

		/* optimize - op size must be consistent (no unestablished IDs) */
		if ((unsigned)m<OPM_IA && f->opt&_Optimize && !(f->flg&_PreEstID))
			m = optimizeMode(m,v);
		if ((r=getOpcode(x|(m<<8)))<0)
			if (m==OPM_IP && (r=getOpcode(x|(OPM_IA<<8)))>=0)
				m = OPM_IA;
			else if ((m==OPM_ZP || m==OPM_AB)
					&& (r=getOpcode(x|(OPM_RL<<8)))>=0)
			{ /* instruction calls for a relative; convert */
				m = OPM_RL, v-= f->org, v-= 2;
				/* check range - not necessarily established in prescan */
				if (v!=(int8)v && !(f->opt&_Prescan))
					{r = _e_range_error; break;}
			} else if (f->opt&_Optimize)
			{ /* last resort: if reduced, check for a larger, alt. mode */
				m = optimizeMode(m,256);
				if ((r=getOpcode(x|(m<<8)))<0)
					{r = _e_op_error; break;} /* no such instruction */
			} else
				{r = _e_op_error; break;}
		if (bw)
			bw(f,r);
		++f->org, f->org&= k, r = 0;
		if (m>=OPM_IA)
			continue;
		if (bw)
			bw(f,v);
		++f->org, f->org&= k;
		if ((unsigned)(m-OPM_IB)<=OPM_AY-OPM_IB)
		{ /* 2-byte mode range */
			if (bw)
				bw(f,v>>8);
			++f->org, f->org&= k;
 			if ((unsigned)v>0xFFFF)
				wrn =~_e_range_error;
		} else if ((unsigned)v>0xFF && v<-0x80 && m!=OPM_RL && x!=OPN_AND)
			wrn =~_e_range_error;
	}
	f->flg&=~_IntoLine;
end:	if (f==&af && af.lab)
		free(af.lab);
	if (_r)
		*_r = r ? r:wrn;
	return s-str;
}

#if __POCC__
 #pragma warn(disable:2154)
#endif

static long deline(char *str,
	long z)
{
	const char	*s = str, *e = s+z;
	char		*d;
	char		c;

	if ((d=str)==NULL || z<=0)
		return 0;
	do
		if ((c=*s++)=='\\' && e-s>1 && IS_EOL(*s))
			++s;
		else
			*d++ = c;
	while(s<e);
//	*d = 0; /* freadline should already buffer for null termination */
	return d-str;
}

/* converts tabs to spaces so the pointer (^) places right */
static void untab(char *s,
	const char *e)
{
	if (!s)
		return;
	while(s<e)
		if (*s++=='\t')
			s[-1] = ' ';
}

static const char	*err_strs[] =
{
	"syntax/param error",
	"memory error",
	"read error",
	"write error",
	"incomplete expression",
	"interrogator already open",
	"ID not found or applicable",
	"operation not supported here",
	"too many/few parentheses",
	"value out of range",
	"comment still open",
	"unterminated string",
	"ID already defined",
	"no such instruction mode",
	"invalid/unsupported label asignment",
	"unknown/unsupported language",
	"unrecognized macro, keyword or instruction",
	"subroutines are always linked, flat",
	"value already set to something else"
};

static const char	*err_strs2[] =
{
	"failed to include file",
	"exceeded maximum include depth",
	"could not open include file",
	"expected 'on' or 'off'",
	"attempt to write to non-code/data space",
	"could not write PRG here",
	"auto-banking wrapped - bank set to #1",
	"no such label",
	"wrap occurred in operation",
	"no endproc before proc",
	"hit end within a proc",
	"hit endproc outside a proc",
	"total name too long"
};

#define MAXIMUM_DEPTH		8

int X6502_AsmFile(FILE *f,
	const char *pfn,
	asm6502frame *af,
	uint8 depth)
{
	char		*str;
	uint32		z, l, x = 1;
	int		r = 0;

	if (!f || !af)
		return _e_general_error;
	str = malloc(0); /* needed for first-empty line(s) case */
	if (depth>=MAXIMUM_DEPTH)
		return _e_include_depth;
	while(z=~0,l=x,(str=freadline(f,&z,str,NULL,&x,'\\'))!=NULL)
	{
		const char	*s = str, *e;
		long		t = deline(str,z);

		t = X6502_Assemble(s,e=s+t,&r,af,pfn);
		if (r==_e_include)
		{
			FILE	*nf = NULL;
			char	*b = NULL;

			s+= t;
			if ((r=X6502_ParseName(s,e,NULL,&b,pfn))>0)
				if ((s=parse_skipwsp(s+r,e))<e && !IS_EOL(*s) && *s!=';')
					r = _e_general_error;
				else
					nf = fopen(b,"r");
			if (nf)
			{
				af->flg&=~_IntoLine; /* for files that immed. start w. a label */
				r = X6502_AsmFile(nf,b,af,depth+1);
				fclose(nf);
			} else if (r>0)
				r = _e_include_open;
			else
				r = r ? r:_e_incomplete_error, t+= r;
			if (b)
				free(b);
			if (r>0) /* error/warning message already handled? */
				goto end;
		}
		if (r)
		{
			long		errc;

		se:	s = NULL;
			if (r>0 && af->opt&_Prescan)
				goto skip;
			if (r==_e_param_separation || r==_e_return)
				r = _e_general_error;
			if ((errc=r)<0)
				errc =~errc;
			if (errc<GET_ARRAY_COUNT(err_strs))
				s = err_strs[errc];
			else if (r<0)
				if ((errc-=~_e_include)<GET_ARRAY_COUNT(err_strs2))
					s = err_strs2[errc];
				else if (!s)
					s = "unknown error";
			if (s)
			{
				untab(str,e);
				if (t>=0 && t<(e-str))
					fprintf(stderr,
"### %s\n"
"### %* ^\n"
"### %s\n"
"File '%s'; Line %u\n\n",
						str,t,s,pfn,l);
				else /* offset returned is meaningless */
					fprintf(stderr,
"# > %s\n"
"##%s%s\n"
"File '%s'; Line %u\n\n",
						str,r<0 ? "# ":" Warning - ",s,pfn,l);
			}
		skip:	if (r<0)
				{r = 1; goto end;}
		} else if (af->flg&_PWrError)
			{r = _e_prg_write_error, t =-1; goto se;}
		if (af->flg&_EndKeyw)
			{r = 0; goto end;}
	}
	r = z ? _e_memory_error:0; /* length - realloc failed */
end:	if (str)
		free(str);
	return r;
}

int X6502_FreeFrame(asm6502frame *f)
{
	int		r = 0;

	if (!f)
		return _e_general_error;
	if (f->lab)
		{free(f->lab); f->lab = NULL, ++r;}
	if (f->mod)
		{free(f->mod); f->mod = NULL, ++r;}
	return r;
}

#ifdef SAMPLE
static long uppow2(long v)
{
	int		x = 1u<<(31-cntlz32(v));

	if (v!=x)
		x+= x;
	return x;
}

typedef struct PRGdata
{
	int32		f, k, bk;
	uint8		b[1];
} PRGdata;

void bwrite(asm6502frame *f,
	uint8 v)
{
	PRGdata		*p;
	int32		a;

	if (!f || (p=f->u)==NULL)
		return;
	if ((a=f->pos++-p->f)<0)
		{f->flg|= _PWrError;}
	if (f->bnk>=0)
		a&= p->bk, a+= f->bnk<<f->bsz;
	p->b[a&p->k] = v;
}

/* this is the "logical minimum" routine */
void bexpand(asm6502frame *f,
	uint8 v)
{
#pragma unused(v)
	PRGdata		*p;
	int32		a;

	if (!f || (p=f->u)==NULL)
		return;
	if ((a=f->pos++-p->f)<0)
		{f->flg|= _PWrError; return;}
	if (a>p->k)
		p->k = a;
}

/* ---- top program */

#define _SyntaxCheck		0x8000
#define _FlatBank		0x4000

typedef struct asm6502params
{
	asm6502frame	af;
	const char	*ofn, *pfn, *cfn;
	long		prz;
	uint16		pfl;
	int16		opt;
	uint8		prgram, prgwram, chrram, chrwram;
} asm6502params;

static int writeOutput(asm6502params *ap)
{
	PRGdata		*p;
	FILE		*f, *cf = NULL;
	long		z;
	int		r = 0, x;

	if (!ap || (p=ap->af.u)==NULL)
		return _e_general_error;
	if (ap->af.map>511)
	{
		fprintf(stderr,"### bad mapper index (%d)\n",
			ap->af.map);
		return 1;
	}
	if ((f=fopen(ap->ofn,"wb"))==NULL)
	{
		fprintf(stderr,"## writeOutput: could not open '%s' for writing.\n",
			ap->ofn);
		return 1;
	}
	z = (p->k+1)>>10;
	if (ap->prz&1023)
		printf("# PRG total: %u bytes / %u KB\n",
			ap->prz,z);
	else
	{
		x = ap->af.map<0 ? 0:16384/ap->prz;
		printf(x>1 ? "# PRG total: %u KB / %u KB x%u\n"
			:"# PRG total: %u KB / %u KB\n",
			ap->prz>>10,z,x);
	}
	if (ap->af.map>=0)
	{ /* iNES */
		static char	h[16] = "NES\x1A";
		long int	b =-1;
		int		ver = 0, mir;
		uint8		wrn = 0;

		if (ap->cfn)
		{
			if ((cf=fopen(ap->cfn,"rb"))==NULL)
			{
				fprintf(stderr,"## writeOutput: could not open '%s' for reading.\n",
					ap->cfn);
				r = 1;
				goto end;
			}
			fseek(cf,0,SEEK_END);
			b = (ftell(cf)+8191)&~8191;
			fseek(cf,0,SEEK_SET);
		}
		x = b>>13;
		if (ap->af.maxchr>=0)
		{
			if (ap->af.maxchr!=x)// && x>=0)
				wrn|= 2;
			x = ap->af.maxchr;
		}
		if (x<0)
			x = 0;
		else
			printf("# CHR total: %u KB / %u KB%s\n",
				x<<13-10,b<0 ? 0:b>>10,wrn ? " (!)":"");
		if ((mir=ap->af.opt&_MirrorMask)==_MirrorMask)
			mir = 0;
		/* set version two if submapper, CHR WRAM or large output size(s) */
		/* CHR RAM is 16KB with mapper 13, 8KB otherwise */
		if ((z=ap->af.maxprg+1)>255 || ap->af.map>255 || x>255
				|| ap->af.opt&_SubMapMask || ap->chrwram
				|| (ap->chrram && ((ap->af.map==13 && ap->chrram>14)
				|| (ap->af.map!=13 && ap->chrram>13))))
			ver = 2<<2;
		h[4] = z; //PRG-low
		h[5] = x; //CHR-low
		/* set the battery bit if PRG-WRAM */
		h[6] = ap->af.map<<4|(mir&1)|(2*!!ap->prgwram); //map-low
		h[7] = (ap->af.map&0xF0)|ver|(mir<<2&8); //map-high & version

		if (ver)
		{
			h[8] = (ap->af.map>>8&1)|(ap->af.opt&_SubMapMask)<<4-_SubMapShift;
			h[9] = (z>>8&0xF)|(x>>4&0xF0); //PRG-high/CHR-high
			h[10] = (ap->prgram ? MAX(ap->prgram,7)-6:0)
				|(ap->prgwram ? MAX(ap->prgwram,7)-6:0)<<4;
			h[11] = (ap->chrram ? MAX(ap->chrram,7)-6:0)
				|(ap->chrwram ? MAX(ap->chrwram,7)-6:0)<<4;
		} else
		{ /* no CHR WRAM; CHR RAM is automatic, and not stored in the header */
			z = MAX(ap->prgram,ap->prgwram);
			h[8] = z ? 1u<<MAX(z,13)-13:0;
		}
		if (fwrite(h,1,16,f)!=16)
			r = _e_write_error;
		x = 16384;
	} else
		x = 0;
	z = 0; /* a loop to repeat output (PRG mirror) */
	do
		if (fwrite(p->b,1,ap->prz,f)!=ap->prz)
			{r = _e_write_error; goto end;}
	while((z+=ap->prz)<x);
	if (cf)
	{ /* CHR data */
		char	b[1024];

		while((z=fread(b,1,sizeof(b),cf))>0)
			if (fwrite(b,1,z,f)!=z)
				{r = _e_write_error; break;}
		fclose(cf);
	}
end:	fclose(f);
	if (r==_e_write_error)
	{
		fprintf(stderr,"### writeOutput: error writing to '%s'.\n",
			ap->ofn);
		r = 1;
	}
	return r;
}

static int asm6502(asm6502params *ap)
{
	FILE		*f;
	int		r;

	if (!ap)
		return _e_general_error;
	if ((f=fopen(ap->pfn,"r"))==NULL)
	{
		fprintf(stderr,"### could not open file '%s'\n",
			ap->pfn);
		return 1;
	}
	if (ap->prz<0)
		ap->prz = 16384; /* default */
	if ((ap->af.u=calloc(1,sizeof(PRGdata)))==NULL) /* keep extra byte */
		return _e_memory_error;

	/* at code point, initially */
	ap->af.bsz = 14;
	ap->af.pos = ((PRGdata*)ap->af.u)->f = ap->pfl;
	ap->af.bnk = ap->af.flg = 0, ap->af.maxprg = ap->af.maxchr =-1;
	ap->af.opt = _Prescan|ap->opt;
	((PRGdata*)ap->af.u)->k = 0; /* writing to that single byte */
	((PRGdata*)ap->af.u)->bk = ap->opt&_FlatBank ? ~0:(1u<<ap->af.bsz)-1;
	ap->af.writef = ap->prz ? bwrite:bexpand;

	if ((r=X6502_AsmFile(f,ap->pfn,&ap->af,0))!=0)
		fprintf(stderr,"### Error(s) prevented further processing.\n");
	else if (f && f!=stdin)
	{
		fseek(f,0,SEEK_SET);

		/* reset for full scan */
		if (ap->af.mod)
			{free(ap->af.mod); ap->af.mod = NULL;}
		ap->af.org = 0, ap->af.pos = ap->pfl, ap->af.bnk = ap->af.flg = 0;
		/* refresh all options - code may change _Optimize, anything else */
		ap->af.opt = ap->opt;
		if (ap->af.map>=0)
		{
			if (!ap->prz)
				ap->prz = 1, fputs("## PRG log. minimum overridden by file.\n",stderr);
			else if (ap->prz>16384)
				ap->prz = (ap->prz+16383)&~16383;
			else /* for PRG mirroring (repeated output in .nes) */
				ap->prz = uppow2(ap->prz);
		} else if (!ap->prz)
		{
			printf("# PRG size set to logical minimum\n");
			if ((ap->prz=((PRGdata*)ap->af.u)->k+1)<=0)
				ap->prz = 1, fputs("## Warning: file wrote no PRG.\n",stderr);
		}
		if (ap->af.maxprg<0)
			ap->af.maxprg = (ap->prz-1)>>ap->af.bsz;
		//ap->prz = (ap->af.maxprg+1)<<ap->af.bsz;

		((PRGdata*)ap->af.u)->k = uppow2(ap->prz)-1;
		if (ap->opt&_SyntaxCheck)
		{
			free(ap->af.u); ap->af.u = NULL;
			ap->af.writef = NULL;
		} else
		{
			PRGdata		*p;

			ap->af.writef = bwrite; /* k (mask) is one less */
			if ((p=realloc(ap->af.u,sizeof(PRGdata)
					+((PRGdata*)ap->af.u)->k))==NULL)
				r = _e_memory_error;
			else
			{
				ap->af.u = p;
				memset(&p->b,0,p->k+1);
			}
		}
		if (!r)
			r = X6502_AsmFile(f,ap->pfn,&ap->af,0);
	}
	X6502_FreeFrame(&ap->af);
	if (f && f!=stdin)
		fclose(f);
	if (ap->af.u)
	{
		if (!ap->ofn)
			ap->ofn = ap->af.map>=0 ? "asm-out.nes":"asm-out.bin";
		if (!r)
			r = writeOutput(ap);
		free(ap->af.u); ap->af.u = NULL;
	} else if (!r)
		printf("# Processing successful.\n");
	return r;
}

/* returns largest negative on error */
static long int parseInt(const char *s)
{
	long int	t;

	return readanyinteger(s,strlen(s),&t) ? t:IMIN(long int);
}

static int parseParams(asm6502params *ap,
	int argc,
	char **argv)
{
	char		**a = argv;
	const char	*s;

	if (!ap)
		return _e_general_error;
	if (argc<=0)
		goto disp;
	ap->pfl = 0x8000, ap->opt = _MirrorMask;
	ap->af.map =-1, ap->prz =-1;
	while(--argc>=0 && (s=*++a)!=NULL)
		if (*s=='-')
		{
			long int	t = 0, mx = 0;

			switch(s[1])
			{
			case 'B': /*ap->opt|= _AutoBank;*/ break;
			case 'C': ap->cfn = s+2;
				if (ap->af.map<0)
					ap->af.map = 0;
				break;
			case 'F': mx = 0xFFFF;
				ap->pfl = t = parseInt(s+2);
				break;
			case 'K': ap->opt|= _SyntaxCheck; break;
			case 'L': ap->opt|= _FlatBank; break;
			case 'M': mx = 511;
				ap->af.map = t = parseInt(s+2);
				break;
			case 'N': ap->ofn = s+2; break;
			case 'O': ap->opt|= _Optimize; break;
			case 'P': if (s[2]=='R')
					{ap->chrram = t = parseInt(s+3), mx = 20; break;}
				else if (s[2]=='W')
					{ap->chrwram = t = parseInt(s+3), mx = 20; break;}
				break;
				mx = 16383, t = parseInt(s+3);
				t = parseInt(s+2), ap->prz = t<<10;
				break;
			case 'R': mx = 2;
				ap->opt&=~_MirrorMask, ap->opt|= t = parseInt(s+2);
				break;
			case 'S': mx = 15;
				ap->opt&=~_SubMapMask, ap->opt|= t = parseInt(s+2)<<_SubMapShift;
				break;
			case 'X': mx = 20, t = parseInt(s+3);
				if (s[2]=='R')
					ap->chrram = t;
				else if (s[2]=='W')
					ap->chrwram = t;
				else
					t = IMIN(t);
				break;

			default:
				fprintf(stderr,"### %s - unknown/unsupported option '%s'.\n",
					*argv,s);
				return 1;
			}
			if ((unsigned)t>mx)
			{
				fprintf(stderr,"### %s - expected integer after %.2s, in [0,%u].\n",
					*argv,s,mx);
				return 1;
			}
		} else
		{
			if (ap->pfn)
			{
				fprintf(stderr,"### %s - one input at a time.\n",
					*argv);
				return 1;
			}
			ap->pfn = s;
		}
	if (!ap->pfn)
	{
		fputs("### no file input\n",stderr);
	disp:	fputs(
"#\n"
//"#  -A           auto-mirror PRG (otherwise cleared)\n"
//"#  -            parse stdin - limited results\n"
//"#  -B           auto-bank (incr. on new ORGs)\n"
"#  -C<name>     load CHR file\n"
"#  -F<value>    code floor (default: $8000)\n"
"#  -K           check syntax only\n"
"#  -L           flat (unrestricted 64KB positioning)\n"
"#  -M<mapper>   (iNES output)\n"
"#  -N<name>     set output name\n"
"#  -O           turn optimization on\n"
"#  -P<value>    set output PRG KB size - 0=packed 32KB\n"
"#  -PR<value>   set PRG RAM - 1<<x bytes (iNES output)\n"
"#  -PW<value>   set PRG WRAM - 1<<x bytes (iNES output)\n"
"#  -R<mirror>   (iNES output) - H=0, V=1, FS=2 \n"
"#  -S<value>    set submapper (iNES output)\n"
"#  -XR<value>   set CHR RAM - 1<<x bytes (iNES output)\n"
"#  -XW<value>   set CHR WRAM - 1<<x bytes (iNES output)\n"
"\n",
			stderr);
		return 1;
	}
	return 0;
}

int main(int argc,
	char **argv)
{
	asm6502params	ap;
	int		r;

	if (argc<=0 || !argv)
		return 0;
	if (!--argc || !argv[1])
		argc = 0, fprintf(stderr,"### usage: %s filename\n",
			*argv);
	memset(&ap,0,sizeof(asm6502params));
	if ((r=parseParams(&ap,argc,argv))==0)
		r = asm6502(&ap);
	if (r<0)
	{
		fprintf(stderr,"### %s - fatal error (%d).\n",
			*argv,r);
		if (r==_e_memory_error)
			fputs("### out of memory.",stderr);
		r = 1;
	}
	return r;
}
#endif
